/* The Great Computer Language Shootout
   http://shootout.alioth.debian.org/
   contributed by Isaac Gouy */
   
   
import java.lang.Math;
import java.text.*;
import java.lang.System;
//import java.lang.management.*;

//var mx = ManagementFactory.getThreadMXBean();
var start = System.currentTimeMillis();
      
var PI = 3.141592653589793;
var SOLAR_MASS = 4 * PI * PI;
var DAYS_PER_YEAR = 365.24;


class Body{
	
   attribute x: double;
   attribute y: double;
   attribute z: double;
   attribute vx: double;
   attribute vy: double;
   attribute vz: double;
   attribute mass: double;
   operation Body(x,y,z,vx,vy,vz,mass): Body;
   operation offsetMomentum(px,py,pz): Body;
   
}


operation Body.Body(x,y,z,vx,vy,vz,mass){
   var self = this;
   
   self.x = x;
   self.y = y;
   self.z = z;
   self.vx = vx;
   self.vy = vy;
   self.vz = vz;
   self.mass = mass;

}


operation Body.offsetMomentum(px,py,pz) {
   var self = this;
   var PI = 3.141592653589793;
   var SOLAR_MASS = 4 * PI * PI;
   
   self.vx = -px / SOLAR_MASS;
   self.vy = -py / SOLAR_MASS;
   self.vz = -pz / SOLAR_MASS;
  
   return this;
}


function Jupiter(){
var PI = 3.141592653589793;
var SOLAR_MASS  = 4 * PI * PI;
var DAYS_PER_YEAR = 365.24;

   return new Body(
      4.84143144246472090e+00,
      -1.16032004402742839e+00,
      -1.03622044471123109e-01,
      1.66007664274403694e-03 * DAYS_PER_YEAR,
      7.69901118419740425e-03 * DAYS_PER_YEAR,
      -6.90460016972063023e-05 * DAYS_PER_YEAR,
      9.54791938424326609e-04 * SOLAR_MASS
   );
}

function Saturn(){
var PI = 3.141592653589793;
var SOLAR_MASS  = 4 * PI * PI;
var DAYS_PER_YEAR = 365.24;
   return new Body(
      8.34336671824457987e+00,
      4.12479856412430479e+00,
      -4.03523417114321381e-01,
      -2.76742510726862411e-03 * DAYS_PER_YEAR,
      4.99852801234917238e-03 * DAYS_PER_YEAR,
      2.30417297573763929e-05 * DAYS_PER_YEAR,
      2.85885980666130812e-04 * SOLAR_MASS
   );
}

function Uranus(){
var PI = 3.141592653589793;
var SOLAR_MASS  = 4 * PI * PI;
var DAYS_PER_YEAR = 365.24;
   return new Body(
      1.28943695621391310e+01,
      -1.51111514016986312e+01,
      -2.23307578892655734e-01,
      2.96460137564761618e-03 * DAYS_PER_YEAR,
      2.37847173959480950e-03 * DAYS_PER_YEAR,
      -2.96589568540237556e-05 * DAYS_PER_YEAR,
      4.36624404335156298e-05 * SOLAR_MASS
   );
}

function Neptune(){
var PI = 3.141592653589793;
var SOLAR_MASS  = 4 * PI * PI;
var DAYS_PER_YEAR = 365.24;
   return new Body(
      1.53796971148509165e+01,
      -2.59193146099879641e+01,
      1.79258772950371181e-01,
      2.68067772490389322e-03 * DAYS_PER_YEAR,
      1.62824170038242295e-03 * DAYS_PER_YEAR,
      -9.51592254519715870e-05 * DAYS_PER_YEAR,
      5.15138902046611451e-05 * SOLAR_MASS
   );
}

function Sun(){
   return new Body(0.0, 0.0, 0.0, 0.0, 0.0, 0.0, SOLAR_MASS);
}



class NBodySystem{

	operation NBodySystem(bodies) : NBodySystem;
	operation advance(dt);
	operation energy(): double;
	attribute bodies: Body*;	
	
}


	
operation NBodySystem.NBodySystem(bodies){
   var self = this;
   self.bodies = bodies;
   var px = 0.0;
   var py = 0.0;
   var pz = 0.0;
   var size = sizeof(self.bodies);
   
   for (i in [0..size]){
      var b = self.bodies[i];
      var m = b.mass;
      px += b.vx * m;
      py += b.vy * m;
      pz += b.vz * m;
       
   }
   self.bodies[0].offsetMomentum(px,py,pz);
}

operation NBodySystem.advance(dt){
   var dx=0;
   var dy=0;
   var dz=0;
   var distance=0;
   var mag=0;
   
   var self = this;
   var size = sizeof(self.bodies);

   for (i in [0..size]) {
      var bodyi = self.bodies[i];
      for (j in [i+1..size]) {
         var bodyj = self.bodies[j];
         dx = bodyi.x - bodyj.x;
         dy = bodyi.y - bodyj.y;
         dz = bodyi.z - bodyj.z;

         distance = Math.sqrt(dx*dx + dy*dy + dz*dz);
         mag = dt / (distance * distance * distance);

         bodyi.vx -= dx * bodyj.mass * mag;
         bodyi.vy -= dy * bodyj.mass * mag;
         bodyi.vz -= dz * bodyj.mass * mag;

         bodyj.vx += dx * bodyi.mass * mag;
         bodyj.vy += dy * bodyi.mass * mag;
         bodyj.vz += dz * bodyi.mass * mag;
      }
   }

   for (i in [0..size]) {
      var body = self.bodies[i];
      body.x += dt * body.vx;
      body.y += dt * body.vy;
      body.z += dt * body.vz;
   }
}

operation NBodySystem.energy(){
   
   var self = this;
   var dx=0;
   var dy=0;
   var dz=0;
   var distance=0;
   var e = 0.0;
   var size = sizeof(self.bodies);

   for (i in [0..size]) {
      var bodyi = self.bodies[i];

      e += 0.5 * bodyi.mass *
         ( bodyi.vx * bodyi.vx
         + bodyi.vy * bodyi.vy
         + bodyi.vz * bodyi.vz );

      for (j in [i+1..size]) {
         var bodyj = self.bodies[j];
         dx = bodyi.x - bodyj.x;
         dy = bodyi.y - bodyj.y;
         dz = bodyi.z - bodyj.z;
				
         distance = Math.sqrt(dx*dx + dy*dy + dz*dz);
         e -= (bodyi.mass * bodyj.mass) / distance;
      }
   }
   return e;
}


var nf = NumberFormat.getInstance();
nf.setMaximumFractionDigits(9);
nf.setMinimumFractionDigits(9);
nf.setGroupingUsed(false);


//var n = arguments[0];
var n = 200000;

var bodies = new NBodySystem( [
   Sun(),Jupiter(),Saturn(),Uranus(),Neptune()
]);


System.out.println(bodies.energy());

for (i in [0..n]){
	bodies.advance(0.01);
	//System.out.println(bodies.energy());
}
System.out.println(bodies.energy());

//var cpu = mx.getCurrentThreadCpuTime()/1000000;
var stop = System.currentTimeMillis();
var end = stop-start;
//System.out.print("cputime/ms:");System.out.println(cpu);
System.out.print("tottime/ms:");System.out.println(end);
//System.out.print("cpu util:");System.out.println(( cpu / (end) * 100));
